/*************************************************************************\
*
*  RPL.C
*
*  This module is responsible for all object loading and unloading from
*  the internal memory of the HP48
*
*  (c)1995 Sebastien Carlier
*
*  Portions (c)2002 by Christoff Giesselink
*
*  Portions (c)2002 by Robert Hildinger
*
\*************************************************************************/

#include <PalmOS.h>
#include "Power48Rsc.h"
#include "Power48BMPRsc.h"
#include <SonyCLIE.h>
#include "emu48.h"

// 48SX | 48GX | 49G  | Name
//#7056A #806E9 #806E9 =TEMPOB
//#7056F #806EE #806EE =TEMPTOP
//#70574 #806F3 #806F3 =RSKTOP   (B)
//#70579 #806F8 #806F8 =DSKTOP   (D1)
//#7066E #807ED #80E9B =AVMEM    (D)
//#705B0 #8072F #8076B =INTRPPTR (D0)

#define TEMPOB   ((optionTarget==HP48SX)?0x7056A:0x806E9)
#define TEMPTOP  ((optionTarget==HP48SX)?0x7056F:0x806EE)
#define RSKTOP   ((optionTarget==HP48SX)?0x70574:0x806F3)
#define DSKTOP   ((optionTarget==HP48SX)?0x70579:0x806F8)
#define AVMEM    ((optionTarget!=HP49G)?((optionTarget==HP48SX)?0x7066E:0x807ED):0x80E9B)
#define INTRPPTR ((optionTarget!=HP49G)?((optionTarget==HP48SX)?0x705B0:0x8072F):0x8076B)

#define objectVFSDirName		"/PALM/PROGRAMS/Power48/Objects/"
#define BINARYHEADER48  		"HPHP48-W"
#define BINARYHEADER49  		"HPHP49-W"

char objectNameTmp[12];

const char mkString[14]={0xd,0x4,0x4,0x4,0x7,0x4,0xb,0x4,0x5,0x4,0x2,0x5,0xa,0x3};
const char mk2String[4]={0xd,0x4,0xb,0x4};
const char mk_230[6]={0x4D, 0x4B, 0x32, 0x2E, 0x33, 0x30};

// search for "MDGKER:MK2.30" or "MDGKER:PREVIE" in port1 of a HP48GX
static Boolean MetakernelPresent(void)
{
	Boolean bMkDetect = false;
	int wVersion;
	
	// card in slot1 of a HP48GX enabled
	if ((optionTarget==HP48GX) && (CARDSTATUS&2))
	{
		// check for Metakernel string "MDGKER:"
		if (!MemCmp(&port1[12],mkString,14))
		{
			bMkDetect = true;				// Metakernel detected
			// check for "MK"
			if (!MemCmp(&port1[26],mk2String,4))
			{
				// get version number
				wVersion = ((port1[30] * 10) + port1[34]) * 10 + port1[36];

				// version newer then V2.30, then compatible with HP OS
				bMkDetect = (wVersion <= 230);
			}
		}
	}
	return bMkDetect;
}


inline char* GetRealAddressNoIO(long d)
{
  	if ((d>=mcRAMBase)&&(d<mcRAMCeil))  // RAM
	{
		return d+(char*)mcRAMDelta;
	}
	else if ((d>=mcCE2Base)&&(d<mcCE2Ceil))
	{
		return d+(char*)mcCE2Delta;
	}
	else if ((d>=mcCE1Base)&&(d<mcCE1Ceil))
	{
		return d+(char*)mcCE1Delta;
	}
  	else   //ROM
  	{
  		if (status49GActive)
  		{
  			d&=0x7ffff;   // 49 ROM is mirrored
  			if (d<0x40000)  // lo bank
  				return flashRomLo+d;
  			else			// hi bank
  				return flashRomHi+d;
  		}
  		else
			return rom+d;
	}
}

inline long Pack5(long addr)  // pack into long
{
	char *n;
	
	n=GetRealAddressNoIO(addr);
	
	return ((long)n[4]<<16)|((long)n[3]<<12)|(n[2]<<8)|(n[1]<<4)|n[0];
}

inline long Pack2(long addr)  // pack into long
{
	char *n;
	
	n=GetRealAddressNoIO(addr);
	
	return (n[1]<<4)|n[0];
}

inline long BufPack5(char *n)  // pack into long
{		
	return ((long)n[4]<<16)|((long)n[3]<<12)|(n[2]<<8)|(n[1]<<4)|n[0];
}

inline long BufPack2(char *n)  // pack into long
{		
	return (n[1]<<4)|n[0];
}

inline void Unpack5(long addr, long val)   // unpack into addr
{
	char *n;
	
	n=GetRealAddressNoIO(addr);
		
	MemSemaphoreReserve(true);
	*n++=val&0xf;
	val>>=4;
	*n++=val&0xf;
	val>>=4;
	*n++=val&0xf;                        
	val>>=4;
	*n++=val&0xf;
	val>>=4;
	*n=val&0xf;
	MemSemaphoreRelease(true);
}

inline void BufUnpack5(char *n, long val)   // unpack into addr
{
	*n++=val&0xf;
	val>>=4;
	*n++=val&0xf;
	val>>=4;
	*n++=val&0xf;                        
	val>>=4;
	*n++=val&0xf;
	val>>=4;
	*n=val&0xf;
}

inline void ReadX(char *buf, long addr, long count)   // store buf from addr
{
	char *n;
	
	n=GetRealAddressNoIO(addr);
	
	do { *buf++=*n++; } while (--count);
}

inline void WriteX(char *buf, long addr, long count)    // move buf into addr  
{
	char *n;
	
	n=GetRealAddressNoIO(addr);
	
	MemSemaphoreReserve(true);
	do { *n++=*buf++; } while (--count);
	MemSemaphoreRelease(true);
}


/*********************************************************************
 * RPL_ObjectName: return the name of the object                     *
 *                 at location 'd' in calc memory                    *
 *********************************************************************/
static char *RPL_ObjectName(long d)
{
	long prolog;
	
	if (d==0)  return "Empty";
	
	prolog = Pack5( d ); // read prolog
	switch (prolog)
	{
		case 0x026AC: return "Flash PTR";
		case 0x02911: return "System Binary";
		case 0x02933: return "Real";
		case 0x02955: return "Long Real";
		case 0x02977: return "Complex";
		case 0x0299D: return "Long Complex";
		case 0x029BF: return "Character";
		case 0x02BAA: return "Extended Pointer";
		case 0x02E92: return "XLIB Name";
		case 0x02686: return "Symbolic matrix";
		case 0x02A74: return "List";
		case 0x02AB8: return "Algebraic";
		case 0x02ADA: return "Unit";
		case 0x02D9D: return "Program";
		case 0x0312B: return "SEMI";
		case 0x02E48: return "Global Name";
		case 0x02E6D: return "Local Name";
		case 0x02AFC: return "Tagged";
		case 0x02A96: return "Directory";
		case 0x02614: return "Precision Integer";
		case 0x026D5: return "Aplet";
		case 0x026FE: return "Mini Font"; 
		case 0x029E8: return "Array";
		case 0x02A0A: return "Linked Array";
		case 0x02A2C: return "String";
		case 0x02A4E: return "Binary Integer";
		case 0x02B1E: return "Graphic";
		case 0x02B40: return "Library";
		case 0x02B62: return "Backup";
		case 0x02B88: return "Library Data";
		case 0x02BCC: return "Reserved 1";
		case 0x02BEE: return "Reserved 2";
		case 0x02C10: return "Reserved 3";
		case 0x02DCC: return "Code";
		case 0x0263A: return "Precision Real";
		case 0x02660: return "Precision Complex";
		default: break;
	}
	StrPrintF(objectNameTmp,"%lx",prolog);
	return objectNameTmp;
}



/*********************************************************************
 * RPL_ObjectSize: return the size (in nibbles) of the object        *
 *                 at location 'd' in calc memory                    *
 *********************************************************************/

static long RPL_ObjectSize(long d)
{
	long n,l = 0;
	long prolog;
	
	prolog = Pack5( d ); // read prolog
	switch (prolog)
	{
		case 0x026AC: l = (optionTarget!=HP49G) ? 5 : 12; break; 					// Flash PTR (HP49G)
		case 0x02911: l = 10; break;												// System Binary
		case 0x02933: l = 21; break; 												// Real
		case 0x02955: l = 26; break; 												// Long Real
		case 0x02977: l = 37; break; 												// Complex
		case 0x0299D: l = 47; break; 												// Long Complex
		case 0x029BF: l =  7; break; 												// Character
		case 0x02BAA: l = 15; break; 												// Extended Pointer
		case 0x02E92: l = 11; break; 												// XLIB Name
		case 0x02686: if (optionTarget!=HP49G) { l = 5; break; }					// Symbolic matrix (HP49G)
		case 0x02A74: 																// List
		case 0x02AB8: 																// Algebraic
		case 0x02ADA: 																// Unit
		case 0x02D9D: n=5; do { l+=n; d+=n; n=RPL_ObjectSize(d); } while (n);		// Program
						   l += 5; break; 
		case 0x0312B: l = 0; break; 												// SEMI
		case 0x02E48: 																// Global Name
		case 0x02E6D: 																// Local Name
		case 0x02AFC: n=7+Pack2(d+5)*2; l=n+RPL_ObjectSize(d+n); break;				// Tagged
		case 0x02A96: 																// Directory
			n = Pack5(d+8);
			if (n==0) // empty dir
			{
				l=13;
			}
			else
			{
				l = 8+n;
				n = Pack2(d+l)*2+4;
				l += n;
				l += RPL_ObjectSize(d+l);
			}
			break;
		case 0x02614:																// Precision Integer (HP49G)
		case 0x026D5: 																// Aplet (HP49G)
		case 0x026FE: if (optionTarget!=HP49G) { l = 5; break; }					// Mini Font (HP49G)
		case 0x029E8: 																// Array
		case 0x02A0A: 																// Linked Array
		case 0x02A2C: 																// String
		case 0x02A4E: 																// Binary Integer
		case 0x02B1E: 																// Graphic
		case 0x02B40: 																// Library
		case 0x02B62: 																// Backup
		case 0x02B88: 																// Library Data
		case 0x02BCC: 																// Reserved 1, Font (HP49G)
		case 0x02BEE: 																// Reserved 2
		case 0x02C10: 																// Reserved 3
		case 0x02DCC: l=5+Pack5(d+5); break;										// Code
		case 0x0263A: l=5; if (optionTarget==HP49G) { l += Pack5(d+l);				// Precision Real (HP49G)
													  l += Pack5(d+l); } break;
		case 0x02660: l=5; if (optionTarget==HP49G) { l += Pack5(d+l);				// Precision Complex (HP49G)
													  l += Pack5(d+l);
													  l += Pack5(d+l);
													  l += Pack5(d+l); } break;	
		default: l=5;
	}
	return l;
}




/**********************************************************************************
 * RPL_ObjectSizeInBuffer: return the size (in nibbles) of the object in          *
 *                         local buffer memory                                    *
 **********************************************************************************/

static long RPL_ObjectSizeInBuffer(char *d)
{
	long n,l = 0;
	long prolog;
	
	prolog = BufPack5( d ); // read prolog
	switch (prolog)
	{
		case 0x026AC: l = (optionTarget!=HP49G) ? 5 : 12; break; 					// Flash PTR (HP49G)
		case 0x02911: l = 10; break;												// System Binary
		case 0x02933: l = 21; break; 												// Real
		case 0x02955: l = 26; break; 												// Long Real
		case 0x02977: l = 37; break; 												// Complex
		case 0x0299D: l = 47; break; 												// Long Complex
		case 0x029BF: l =  7; break; 												// Character
		case 0x02BAA: l = 15; break; 												// Extended Pointer
		case 0x02E92: l = 11; break; 												// XLIB Name
		case 0x02686: if (optionTarget!=HP49G) { l = 5; break; }					// Symbolic matrix (HP49G)
		case 0x02A74: 																// List
		case 0x02AB8: 																// Algebraic
		case 0x02ADA: 																// Unit
		case 0x02D9D: n=5;do {l+=n;d+=n;n=RPL_ObjectSizeInBuffer(d);} while (n);	// Program
						   l += 5; break; 
		case 0x0312B: l = 0; break; 												// SEMI
		case 0x02E48: 																// Global Name
		case 0x02E6D: 																// Local Name
		case 0x02AFC: n=7+BufPack2(d+5)*2; l=n+RPL_ObjectSizeInBuffer(d+n); break;	// Tagged
		case 0x02A96: 																// Directory
			n = BufPack5(d+8);
			if (n==0) // empty dir
			{
				l=13;
			}
			else
			{
				l = 8+n;
				n = BufPack2(d+l)*2+4;
				l += n;
				l += RPL_ObjectSizeInBuffer(d+l);
			}
			break;
		case 0x02614:																// Precision Integer (HP49G)
		case 0x026D5: 																// Aplet (HP49G)
		case 0x026FE: if (optionTarget!=HP49G) { l = 5; break; }					// Mini Font (HP49G)
		case 0x029E8: 																// Array
		case 0x02A0A: 																// Linked Array
		case 0x02A2C: 																// String
		case 0x02A4E: 																// Binary Integer
		case 0x02B1E: 																// Graphic
		case 0x02B40: 																// Library
		case 0x02B62: 																// Backup
		case 0x02B88: 																// Library Data
		case 0x02BCC: 																// Reserved 1, Font (HP49G)
		case 0x02BEE: 																// Reserved 2
		case 0x02C10: 																// Reserved 3
		case 0x02DCC: l=5+BufPack5(d+5); break;										// Code
		case 0x0263A: l=5; if (optionTarget==HP49G) { l += BufPack5(d+l);			// Precision Real (HP49G)
													  l += BufPack5(d+l); } break;
		case 0x02660: l=5; if (optionTarget==HP49G) { l += BufPack5(d+l);			// Precision Complex (HP49G)
													  l += BufPack5(d+l);
													  l += BufPack5(d+l);
													  l += BufPack5(d+l); } break;	
		default: l=5;
	}
	return l;
}




/*********************************************************************
 * RPL_SkipOb: return the address after skiping object in 'd'        *
 *********************************************************************/
static long RPL_SkipOb( long d )
{
   return( d + RPL_ObjectSize(d) );
}



/*********************************************************************
 * RPL_CreateTemp: return address to 'l' free nibs in Temp Memory    *
 *********************************************************************/
static long RPL_CreateTemp(long l)
{
	long a, b, c;
	char *p;
	
	l += 6;									// memory for link field (5) + marker (1) and end
	a = Pack5(TEMPTOP);						// tail address of top object 
	b = Pack5(RSKTOP);						// tail address of rtn stack
	c = Pack5(DSKTOP);						// top of data stack
	
	if ((b+l)>c) return 0;					// check if there's enough memory to move DSKTOP

	Unpack5(TEMPTOP, a+l);					// adjust new end of top object
	Unpack5(RSKTOP,  b+l);					// adjust new end of rtn stack
	Unpack5(AVMEM, (c-b-l)/5);				// calculate free memory (*5 nibbles)

	p = (char*)MemPtrNew(b-a);				// move down rtn stack
	ReadX(p,a,b-a);
	WriteX(p,a+l,b-a);
	MemPtrFree(p);
	
	
	Unpack5(a+l-5,l);						// set object length field
	return (a+1);							// return base address of new object
}




/*********************************************************************
 * RPL_Pick: get the address of the object in stack level 'l'        *
 *********************************************************************/
static long RPL_Pick(int l)
{
	long stkp;

	if (!l) return 0;
	if (MetakernelPresent()) ++l;					// Metakernel support
	stkp = Pack5(DSKTOP) + (l-1)*5;
	return Pack5(stkp);
}



/*********************************************************************
 * RPL_Push: push the object in n into stack level 1                 *
 *********************************************************************/
static void RPL_Push(long n)
{
	long stkp, avmem;

	avmem = Pack5(AVMEM);					// amount of free memory
	if (avmem==0) return;					// no memory free
	avmem--;								// fetch memory
	Unpack5( AVMEM, avmem );					// save new amount of free memory
	stkp = Pack5( DSKTOP );					// get pointer to stack level 1
	if (MetakernelPresent())				// Metakernel running ?
	{
		Unpack5(stkp-5,Pack5(stkp));			// copy object pointer of stack level 1 to new stack level 1 entry
		Unpack5(stkp,n);						// save pointer to new object on stack level 2
		stkp-=5;							// fetch new stack entry
	}
	else
	{
		stkp-=5;							// fetch new stack entry
		Unpack5( stkp, n );					// save pointer to new object on stack level 1
	}
	Unpack5( DSKTOP, stkp );					// save new pointer to stack level 1
}



/*********************************************************************
 * LoadObject: Save the object in level1 to a VFS file               *
 *********************************************************************/

static Boolean SaveObject(char *fileName)
{
	UInt16		volRefNum;
	UInt32		volIterator = vfsIteratorStart;
	FileRef		myRef,dirRef;
	UInt8		*buf;
	void		*mem;
	UInt32 		i;
	long		objAddress,objSize=0;
	char		fileNameFull[120];
	Err			error;
	Boolean		dirFound=false;
	char*		calcMem;
	
	StrPrintF(fileNameFull,objectVFSDirName);
	StrCat(fileNameFull,fileName);
	
	objAddress = RPL_Pick(1);
	if (objAddress == 0) return false;
	objSize = (RPL_SkipOb(objAddress) - objAddress + 1) / 2;

	if (FtrPtrNew(appFileCreator, tempObjMemFtrNum, objSize, &mem)) return false;
	
	buf=(UInt8*)mem;
	
	MemSemaphoreReserve(true);
	calcMem=GetRealAddressNoIO(objAddress);
	for( i = 0; i < objSize; i++ )
	{
		buf[i]=calcMem[i*2]|(calcMem[(i*2)+1]<<4);
	}
	MemSemaphoreRelease(true);
	
	while (volIterator != vfsIteratorStop)
	{
		if (VFSVolumeEnumerate(&volRefNum, &volIterator)) return false;
		error = VFSFileOpen(volRefNum,objectVFSDirName,vfsModeRead,&dirRef);
		if (!error) 
		{
			dirFound=true;
			break;
		}
	}
	if (!dirFound) 
	{
		FtrPtrFree(appFileCreator, tempObjMemFtrNum);
		return false;
	}
	
	VFSFileClose(dirRef);
	
	error = VFSFileCreate(volRefNum,fileNameFull);
	if (error==vfsErrFileAlreadyExists)
	{
		VFSFileDelete(volRefNum,fileNameFull);
		VFSFileCreate(volRefNum,fileNameFull);
	}
	
	if (VFSFileOpen(volRefNum,fileNameFull,vfsModeWrite,&myRef))
	{
		FtrPtrFree(appFileCreator, tempObjMemFtrNum);
		return false;
	}

	if (optionTarget!=HP49G)
		VFSFileWrite(myRef,8,BINARYHEADER48,NULL);
	else
		VFSFileWrite(myRef,8,BINARYHEADER49,NULL);
	
	VFSFileWrite(myRef,objSize,buf,NULL);
	
	VFSFileClose(myRef);

	FtrPtrFree(appFileCreator, tempObjMemFtrNum);
	return true;
}



/*********************************************************************
 * LoadObject: load an HP48 object or an ascii file on the stack     *
 *********************************************************************/
Boolean LoadObject(char *fileName)
{
	UInt16		volRefNum;
	UInt32		volIterator = vfsIteratorStart;
	FileRef		myRef;
	UInt32		fileSize;
	UInt8		*buf;
	void		*mem;
	UInt32 		numRead;
	Boolean		objectIsBinary;
	UInt32 		i;
	UInt8		byte; 
	long		objAddress,objSize=0;
	char		head[5];	
	char		fileNameFull[120];
	Err			error;
	Boolean		fileFound=false;
	
	StrPrintF(fileNameFull,objectVFSDirName);
	StrCat(fileNameFull,fileName);
	
	while (volIterator != vfsIteratorStop)
	{
		if (VFSVolumeEnumerate(&volRefNum, &volIterator)) return false;

		// check for presence of file
		error = VFSFileOpen(volRefNum,fileNameFull,vfsModeRead,&myRef);
		if (!error) 
		{
			fileFound=true;
			break;
		}
	}

	if (!fileFound) return false;
	
	if (VFSFileSize(myRef,&fileSize)) return false;
	
	if (fileSize>0x40000)				// max files size 256K
		return false;	
		
	if (FtrPtrNew(appFileCreator, tempObjMemFtrNum, fileSize*2, &mem)) return false;
	
	buf=(UInt8*)mem;
	
	MemSemaphoreReserve(true);
	
 	VFSFileRead(myRef,fileSize,&buf[fileSize],&numRead);
	VFSFileClose(myRef);
 	if (numRead != fileSize )
	{
		MemSemaphoreRelease(true);
		FtrPtrFree(appFileCreator, tempObjMemFtrNum);
		return false;
	}
	
	// MK 2.30 CRC patch start
	// this nasty bit of code patches out a CRC check in MetaKernel 2.30
	// allowing it to run correctly on this emulator. This is a temporary solution
	// until I can think of something better.
	if (fileSize>0x1498f)
	{
		if (!MemCmp(&buf[fileSize+26],mk_230,6))  // Are we loading MK 2.30
		{
			if ((buf[fileSize+0x1498d]==0x84) && (buf[fileSize+0x1498e]==0x40) && (buf[fileSize+0x1498f]==0x01))
			{ 
				buf[fileSize+0x1498d]=0x04;  // patch code
				buf[fileSize+0x1498e]=0x00;
				buf[fileSize+0x1498f]=0x00;
			}
		}
	}
	// MK 2.30 CRC patch end
	
	objectIsBinary = ((buf[fileSize+0]==0x48) &&
					  (buf[fileSize+1]==0x50) &&
					  (buf[fileSize+2]==0x48) &&
					  (buf[fileSize+3]==0x50) &&
					  (buf[fileSize+4]==0x34) &&
					  (buf[fileSize+5]==((optionTarget!=HP49G) ? 0x38 : 0x39)) &&
					  (buf[fileSize+6]==0x2d));

	for( i = 0; i < fileSize; i++ )
	{
		byte=buf[i+fileSize];
		buf[i*2]=byte&0xf;
		buf[(i*2)+1]=byte>>4;
	}

	MemSemaphoreRelease(true);
		
	if (objectIsBinary)
	{ // load as binary
		objSize    = RPL_ObjectSizeInBuffer((char*)buf+16);
		objAddress = RPL_CreateTemp(objSize);
		
		if (objAddress == 0) 
		{
			FtrPtrFree(appFileCreator, tempObjMemFtrNum);
			return false;
		}

		WriteX((char*)buf+16,objAddress,objSize);
	}
	else
	{ // load as string
		objSize = fileSize*2;
		objAddress = RPL_CreateTemp(objSize+10);
		if (objAddress == 0)
		{
			FtrPtrFree(appFileCreator, tempObjMemFtrNum);
			return false;
		}

		BufUnpack5(head,0x02A2C);			// String
		WriteX(head,objAddress,5);
		BufUnpack5(head,objSize+5);			// length of String
		WriteX(head,objAddress+5,5);
		WriteX((char*)buf,objAddress+10,objSize);	// data
	}
	RPL_Push(objAddress);
	
	FtrPtrFree(appFileCreator, tempObjMemFtrNum);
	return true;
}





#define objectLSDialogOriginX	19
#define objectLSDialogOriginY	19
#define objectLSDialogExtentX	282
#define objectLSDialogExtentY	262

Char **fileList;
UInt16 *fileSizeList;
Int16 listPosition;
Int16 numFiles;
Int16 selectedFile=0;
Boolean selectionHighlighted=false;

const objectButton objectLSLayout[] =  // coordinates relative to upper left of objectLS dialog
{
	{ 131,   3, 203,  31, 2},		// objectLoad, scope field def: 1 = active in Load dialog
	{ 206,   3, 279,  31, 1},		// objectSave,                  2 = active in Save dialog
	{  19, 219,  88, 253, 3},		// objectCancel,                3 = active in both
	{ 193, 219, 262, 253, 3},		// objectOK
	{   7,  58, 233, 200, 1},		// objectList
	{ 244, 150, 270, 173, 1},		// objectUp
	{ 244, 179, 270, 202, 1},		// objectDown
}; 


static UInt16 TranslatePenToObjectButton(Int16 x, Int16 y)
{
	UInt16 i;
	
	for (i=0;i<maxObjectButtons;i++)
	{
		if ((x >= ((objectLSLayout[i].upperLeftX+objectLSDialogOriginX)>>1)) &&
		    (x <= ((objectLSLayout[i].lowerRightX+objectLSDialogOriginX)>>1)) &&
		    (y >= ((objectLSLayout[i].upperLeftY+objectLSDialogOriginY)>>1)) &&
		    (y <= ((objectLSLayout[i].lowerRightY+objectLSDialogOriginY)>>1)) &&
		    (statusObjectLSActive&objectLSLayout[i].scope))
		{
			return i;
		}
	}
	return noObject;
}


static void UpdateObjectFileListDisplay(int actionCode)
{
	RectangleType myRect,vacatedRect;
	Int16 index;
	Int16 maxFiles;
	Int16 position=58;
	RGBColorType prevRGB,prevRGB2;
	char tmp[20];
	
	maxFiles=(((numFiles-listPosition)<13)?(numFiles-listPosition):13);
	HRFntSetFont(hrLibRefNum, hrTinyFont);
	WinSetTextColorRGB(&blackRGB,NULL);
	WinSetForeColorRGB(&greyRGB,&prevRGB);
	WinSetBackColorRGB(&greyRGB,&prevRGB2);

	myRect.topLeft.x = objectLSLayout[objectList].upperLeftX+objectLSDialogOriginX;
	myRect.topLeft.y = objectLSLayout[objectList].upperLeftY+objectLSDialogOriginY;
	myRect.extent.x = objectLSLayout[objectList].lowerRightX-objectLSLayout[objectList].upperLeftX+1;
	myRect.extent.y = objectLSLayout[objectList].lowerRightY-objectLSLayout[objectList].upperLeftY+1;
	switch (actionCode)
	{
		case 1:
			HRWinScrollRectangle(hrLibRefNum, &myRect, winUp, 11, &vacatedRect);
			HRWinDrawRectangle(hrLibRefNum, &vacatedRect, 0);
			HRWinDrawTruncChars(hrLibRefNum, fileList[listPosition+12], StrLen(fileList[listPosition+12]), 10+objectLSDialogOriginX, position+objectLSDialogOriginY+132, 184);
			StrPrintF(tmp,"%dK",fileSizeList[listPosition+12]);
			HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 196+objectLSDialogOriginX, position+objectLSDialogOriginY+132, 28);
			break;
		case 2:
			HRWinScrollRectangle(hrLibRefNum, &myRect, winDown, 11, &vacatedRect);
			HRWinDrawRectangle(hrLibRefNum, &vacatedRect, 0);
			HRWinDrawTruncChars(hrLibRefNum, fileList[listPosition], StrLen(fileList[listPosition]), 10+objectLSDialogOriginX, position+objectLSDialogOriginY, 184);
			StrPrintF(tmp,"%dK",fileSizeList[listPosition]);
			HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 196+objectLSDialogOriginX, position+objectLSDialogOriginY, 28);
			break;
		default:
			HRWinDrawRectangle(hrLibRefNum, &myRect, 0);
			for (index=listPosition;index<(maxFiles+listPosition);index++)
			{
				HRWinDrawTruncChars(hrLibRefNum, fileList[index], StrLen(fileList[index]), 10+objectLSDialogOriginX, position+objectLSDialogOriginY, 184);
				StrPrintF(tmp,"%dK",fileSizeList[index]);
				HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 196+objectLSDialogOriginX, position+objectLSDialogOriginY, 28);
				position+=11;
			}
			break;
	}
	
	if ((selectedFile>=listPosition)&&(selectedFile<=(listPosition+12)))
	{
		if (!selectionHighlighted)
		{
			WinSetForeColorRGB(&blackRGB,NULL);
			WinSetDrawMode(winSwap);
			myRect.topLeft.x = objectLSLayout[objectList].upperLeftX+objectLSDialogOriginX;
			myRect.topLeft.y = objectLSLayout[objectList].upperLeftY+objectLSDialogOriginY+((selectedFile-listPosition)*11);
			myRect.extent.x = objectLSLayout[objectList].lowerRightX-objectLSLayout[objectList].upperLeftX+1;
			myRect.extent.y = 11;
			HRWinPaintRectangle(hrLibRefNum,&myRect,0);
			WinSetDrawMode(winPaint);
			selectionHighlighted=true;
		}
	}
	else
	{
		selectionHighlighted=false;
	}	
	WinSetBackColorRGB(&prevRGB2,NULL);
	WinSetForeColorRGB(&prevRGB,NULL);
}


static void DepressObjectLSButton(UInt16 objectCode)
{
	RectangleType 	myRect;
	RGBColorType  	prevRGB;
	Int16			penX,penY;
	Boolean			penDown=false;
	Int16			count=0;
	Int16			tempPos;
	Int16			prevSelectedFile=selectedFile;
	
	WinSetForeColorRGB(&redRGB,&prevRGB);

	switch (objectCode)
	{
		case objectUp:
		case objectDown:
			WinSetBackColorRGB(&blackRGB,NULL);
			WinSetDrawMode(winSwap);
			myRect.topLeft.x = objectLSLayout[objectCode].upperLeftX+objectLSDialogOriginX;
			myRect.topLeft.y = objectLSLayout[objectCode].upperLeftY+objectLSDialogOriginY;
			myRect.extent.x = objectLSLayout[objectCode].lowerRightX-objectLSLayout[objectCode].upperLeftX+1;
			myRect.extent.y = objectLSLayout[objectCode].lowerRightY-objectLSLayout[objectCode].upperLeftY+1;
			HRWinPaintRectangle(hrLibRefNum,&myRect,0);
			do 
			{
				EvtGetPen(&penX,&penY,&penDown);
				if (!count)
				{
					count = 100;
					if (objectCode==objectDown) 
					{
						if (listPosition<(numFiles-13))
						{
							listPosition++;
							UpdateObjectFileListDisplay(1);
						}
					}
					if (objectCode==objectUp) 
					{
						if (listPosition>0)
						{
							listPosition--;
							UpdateObjectFileListDisplay(2);
						}
					}
				}
				count--;
			} while (penDown);
			WinSetBackColorRGB(&blackRGB,NULL);
			WinSetDrawMode(winSwap);
			HRWinPaintRectangle(hrLibRefNum,&myRect,0);
			WinSetDrawMode(winPaint);
			WinSetBackColorRGB(&backRGB,NULL);
			break;
		case objectList:
			do 
			{
				EvtGetPen(&penX,&penY,&penDown);
				tempPos = ((Int16)((penY*2)-(objectLSLayout[objectList].upperLeftY+objectLSDialogOriginY))/11);
				if (tempPos<0) tempPos=0;
				if (tempPos>(numFiles-listPosition-1)) tempPos=(numFiles-listPosition-1);
				if (tempPos>12) tempPos=12;
				selectedFile=listPosition+tempPos;
				if (selectedFile!=prevSelectedFile)
				{
					selectionHighlighted=false;
					UpdateObjectFileListDisplay(0);
					prevSelectedFile=selectedFile;
				}
			} while (penDown);
			break;
		case objectLoad:
		case objectSave:
		case objectCancel:
		case objectOK:
			WinSetBackColorRGB(&greyRGB,NULL);
			WinSetDrawMode(winSwap);
			myRect.topLeft.x = objectLSLayout[objectCode].upperLeftX+objectLSDialogOriginX;
			myRect.topLeft.y = objectLSLayout[objectCode].upperLeftY+objectLSDialogOriginY;
			myRect.extent.x = objectLSLayout[objectCode].lowerRightX-objectLSLayout[objectCode].upperLeftX+1;
			myRect.extent.y = objectLSLayout[objectCode].lowerRightY-objectLSLayout[objectCode].upperLeftY+1;
			HRWinPaintRectangle(hrLibRefNum,&myRect,10);
			WinSetDrawMode(winPaint);
			WinSetBackColorRGB(&backRGB,NULL);
			break;
		default:break;
	}
	
	WinSetForeColorRGB(&prevRGB,NULL);
}


static void ReleaseObjectLSButton(UInt16 objectCode)
{
	RectangleType myRect;
	RGBColorType  prevRGB;
	
	WinSetForeColorRGB(&redRGB,&prevRGB);

	switch (objectCode)
	{
		case objectLoad:
		case objectSave:
		case objectCancel:
		case objectOK:
			WinSetBackColorRGB(&greyRGB,NULL);
			WinSetDrawMode(winSwap);
			myRect.topLeft.x = objectLSLayout[objectCode].upperLeftX+objectLSDialogOriginX;
			myRect.topLeft.y = objectLSLayout[objectCode].upperLeftY+objectLSDialogOriginY;
			myRect.extent.x = objectLSLayout[objectCode].lowerRightX-objectLSLayout[objectCode].upperLeftX+1;
			myRect.extent.y = objectLSLayout[objectCode].lowerRightY-objectLSLayout[objectCode].upperLeftY+1;
			HRWinPaintRectangle(hrLibRefNum,&myRect,10);
			WinSetDrawMode(winPaint);
			break;
		default:break;
	}
	
	WinSetForeColorRGB(&prevRGB,NULL);
	WinSetBackColorRGB(&backRGB,NULL);
}


static void UpdateAvailableMemoryDisplay(void)
{
	RGBColorType prevRGB;
	Char tmp[10];
	long b, c;
	int freeMem;
	
	b = Pack5(RSKTOP);				// tail address of rtn stack
	c = Pack5(DSKTOP);				// top of data stack
	
	freeMem = (int)((c-b)/2048);	// free memory in kilobytes
	
	HRFntSetFont(hrLibRefNum, hrSmallFont);
	WinSetTextColorRGB(&blackRGB,NULL);
	WinSetForeColorRGB(&greyRGB,&prevRGB);
	WinSetBackColorRGB(&greyRGB,NULL);

	StrPrintF(tmp,"%dK",freeMem);
	HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 240+objectLSDialogOriginX, 96+objectLSDialogOriginY, 36);

	WinSetForeColorRGB(&prevRGB,NULL);
}


static void UpdateStackInfoDisplay(void)
{
	RGBColorType prevRGB;
	Char tmp[40];
	
	HRFntSetFont(hrLibRefNum, hrSmallFont);
	WinSetTextColorRGB(&blackRGB,NULL);
	WinSetForeColorRGB(&greyRGB,&prevRGB);
	WinSetBackColorRGB(&greyRGB,NULL);

	StrPrintF(tmp,"Level 4: ");
	StrCat(tmp,RPL_ObjectName(RPL_Pick(4)));
	HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 50+objectLSDialogOriginX, 60+objectLSDialogOriginY, 200);
	StrPrintF(tmp,"Level 3: ");
	StrCat(tmp,RPL_ObjectName(RPL_Pick(3)));
	HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 50+objectLSDialogOriginX, 75+objectLSDialogOriginY, 200);
	StrPrintF(tmp,"Level 2: ");
	StrCat(tmp,RPL_ObjectName(RPL_Pick(2)));
	HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 50+objectLSDialogOriginX, 90+objectLSDialogOriginY, 200);
	StrPrintF(tmp,"Level 1: ");
	StrCat(tmp,RPL_ObjectName(RPL_Pick(1)));
	HRWinDrawTruncChars(hrLibRefNum, tmp, StrLen(tmp), 50+objectLSDialogOriginX, 105+objectLSDialogOriginY, 200);

	WinSetForeColorRGB(&prevRGB,NULL);
}


static void UpdateObjectLoadOrSaveTab(UInt16 saveFieldIndex, Boolean firstPass)
{
	MemHandle 		picHandle;
	BitmapPtr 		picBitmap;
	FieldType 		*saveFileField;
	FormPtr 		frmP;
	WinHandle		lsBackStore;
	Coord 			picWidth,picHeight;
	Err				error;
	
	frmP = FrmGetActiveForm();
	saveFileField = FrmGetObjectPtr(frmP, saveFieldIndex);

	if (statusObjectLSActive==1)  // make sure file name field doesn't show for load tab
	{
		FldReleaseFocus(saveFileField);     // use FldSetUsable & FldReleaseFocus instead of FrmHideObject
		FldSetUsable(saveFileField,false);  // I don't like how FrmHideObject erases...
	}
	
	picHandle = DmGetResource(bitmapRsc, ((statusObjectLSActive==1)?ObjectLoadDialogBitmapFamily:ObjectSaveDialogBitmapFamily));
	picBitmap = MemHandleLock(picHandle);
	BmpGetDimensions(picBitmap,&picWidth,&picHeight,NULL);
	lsBackStore = HRWinCreateOffscreenWindow(hrLibRefNum, 320,picHeight,screenFormat,&error);
	if (error) 
	{
		WinSetDrawWindow(mainDisplayWindow);
		FastDrawBitmap(picBitmap,objectLSDialogOriginX,objectLSDialogOriginY);
	}
	else
	{
		WinSetDrawWindow(lsBackStore);
		FastDrawBitmap(picBitmap,objectLSDialogOriginX,0);
		WinSetDrawWindow(mainDisplayWindow);
		PixelDissolve(lsBackStore,mainDisplayWindow,320L*(UInt32)picHeight,0,objectLSDialogOriginY,true);
		WinDeleteWindow(lsBackStore,false);
	}
	MemHandleUnlock(picHandle);
	DmReleaseResource(picHandle);
	

	if (statusObjectLSActive==2)  // show file name field for save tab
	{
		if (!firstPass)
		{
			FrmShowObject(frmP,saveFieldIndex);
			FrmSetFocus(frmP,saveFieldIndex);
			FldDrawField(saveFileField);
			UpdateStackInfoDisplay();
		}
	}
	else
	{	
		selectionHighlighted=false;
		if (!firstPass)
			UpdateObjectFileListDisplay(0);
		UpdateAvailableMemoryDisplay();
	}
}



/*************************************************************************\
*
*  FUNCTION:  OpenObjectLoadSaveDialog(void)
*
\*************************************************************************/
void OpenObjectLoadSaveDialog(void)
{
	RectangleType 	oldRect,opRect;
    EventType 		event;
	Int16 			penX, penY;
	Boolean 		penDown;
	RGBColorType 	prevRGB,prevRGB2;
	FieldType 		*saveFileField;
	FormPtr 		frmP;
	MemHandle 		saveFileHandle;
	MemPtr			saveFilePtr;
	UInt16 			saveFieldIndex;
	UInt16			objectCode;
	UInt16			objectReleasePending=noObject;
	int				applyChangesOrCancel=0;
	Err				error;
	FileRef			dirRef,fileRef;
	Boolean			dirFound=false;
	UInt16			volRefNum;
	UInt32			volIterator = vfsIteratorStart;
	UInt32			fileIterator = vfsIteratorStart;
	FileInfoType	fileInfo;
	Char 			fileName[256];
	Char			*tempFileNamePtr;
	Char			*dirFileNames[200];  // set a maximum of 200 files displayed
	UInt16			dirFileSizes[200];
	Int16			i,j;
	int				fileIndex=0;
	UInt32			fileSize;
	Char			*temp;
	
	if (fileName==NULL) return;
	
	listPosition=0;
	
	opRect.topLeft.x=0;
	opRect.topLeft.y=0;
	opRect.extent.x=320;
	opRect.extent.y=objectLSDialogOriginY+objectLSDialogExtentY;
	HRWinCopyRectangle(hrLibRefNum,mainDisplayWindow,backStoreWindow,&opRect,opRect.topLeft.x,opRect.topLeft.y,winPaint);

	UIColorSetTableEntry(UIFieldBackground,&greyRGB);
	UIColorSetTableEntry(UIFieldText,&blackRGB);
	UIColorSetTableEntry(UIFieldTextLines,&redRGB);
	UIColorSetTableEntry(UIFieldCaret,&redRGB);
	UIColorSetTableEntry(UIFieldTextHighlightBackground,&redRGB);
	UIColorSetTableEntry(UIFieldTextHighlightForeground,&blackRGB);

    saveFileHandle = MemHandleNew(40);

	frmP = FrmGetActiveForm();
	saveFieldIndex = FrmGetObjectIndex(frmP,MainSaveFileField);
	saveFileField = FrmGetObjectPtr(frmP, saveFieldIndex);

	UpdateObjectLoadOrSaveTab(saveFieldIndex,true);
	
	// Print message saying "please wait"...
	HRFntSetFont(hrLibRefNum, hrSmallFont);
	WinSetTextColorRGB(&redRGB,NULL);
	WinSetForeColorRGB(&greyRGB,&prevRGB);
	WinSetBackColorRGB(&greyRGB,&prevRGB2);
	HRWinDrawTruncChars(hrLibRefNum, "Getting files, please wait...", 29, 20+objectLSDialogOriginX, 70+objectLSDialogOriginY, 200);

	while (volIterator != vfsIteratorStop)
	{
		if (VFSVolumeEnumerate(&volRefNum, &volIterator)) goto exitObjectLS;
		error = VFSFileOpen(volRefNum,objectVFSDirName,vfsModeRead,&dirRef);
		if (!error) 
		{
			dirFound=true;
			break;
		}
	}
	if (!dirFound) 
	{
		OpenAlertConfirmDialog(alert,noVFSDirectoryString,true,false);
		goto exitObjectLS;
	}
		
	fileInfo.nameP = fileName;
	fileInfo.nameBufLen = sizeof(fileName);
	
	while (fileIterator != vfsIteratorStop)
	{
		if (VFSDirEntryEnumerate(dirRef, &fileIterator, &fileInfo)) break;
		tempFileNamePtr = MemPtrNew(StrLen(fileInfo.nameP)+1);
		StrCopy(tempFileNamePtr,fileInfo.nameP);
		dirFileNames[fileIndex] = tempFileNamePtr;
		fileIndex++;
	}
	numFiles=fileIndex;
	fileList=dirFileNames;
	VFSFileClose(dirRef);

    saveFilePtr = MemHandleLock(saveFileHandle);
	for (i=0;i<100;i++)
	{
		StrPrintF(saveFilePtr,"Object%d.p48",i);
		for (j=0;j<numFiles;j++)
		{
			if (StrCaselessCompare(saveFilePtr,dirFileNames[j])==0) break;
		}
		if (j==numFiles) break;
	}
	if (i==100) StrPrintF(saveFilePtr,"");
    MemHandleUnlock(saveFileHandle);
	FldSetTextHandle( saveFileField, saveFileHandle );
    
	HRWinDrawTruncChars(hrLibRefNum, "Sorting files, please wait...   ", 29, 20+objectLSDialogOriginX, 70+objectLSDialogOriginY, 200);
	for (i=0;i<numFiles-1;i++)
	{
		for (j=0;j<numFiles-1-i;j++)
		{
			if (StrCaselessCompare(dirFileNames[j+1],dirFileNames[j])<0)
			{
				temp = dirFileNames[j];
				dirFileNames[j] = dirFileNames[j+1];
				dirFileNames[j+1] = temp;
			}
		}
	}

	HRWinDrawTruncChars(hrLibRefNum, "Getting sizes, please wait...   ", 32, 20+objectLSDialogOriginX, 70+objectLSDialogOriginY, 200);
	
	for (i=0;i<numFiles;i++)
	{
		dirFileSizes[i]=0;
		StrPrintF(fileName,objectVFSDirName);
		StrCat(fileName,dirFileNames[i]);
		if (VFSFileOpen(volRefNum,fileName,vfsModeRead,&fileRef)) continue;
		if (VFSFileSize(fileRef,&fileSize)) continue;
		VFSFileClose(fileRef);
		dirFileSizes[i]=(UInt16)(fileSize/1024);
	}
	fileSizeList=dirFileSizes;

	oldRect.topLeft.x = 19+objectLSDialogOriginX;
	oldRect.topLeft.y = 69+objectLSDialogOriginY;
	oldRect.extent.x  = 202;
	oldRect.extent.y  = 13;
	HRWinDrawRectangle(hrLibRefNum,&oldRect,0);
	WinSetTextColorRGB(&blackRGB,NULL);
	WinSetForeColorRGB(&prevRGB,NULL);
	WinSetBackColorRGB(&prevRGB2,NULL);
	
	if (statusObjectLSActive==1)
	{
		UpdateObjectFileListDisplay(0);
	}
	else
	{
		FrmShowObject(frmP,saveFieldIndex);
		FldGrabFocus(saveFileField);
		FldDrawField(saveFileField);
		UpdateStackInfoDisplay();
	}

UserInputLoop:

	EvtFlushPenQueue();

	if (!option320x320)
	{
		SilkLibResizeDispWin(silkLibRefNum, silkResizeNormal);  // bring up graffiti area
		SilkLibDisableResize(silkLibRefNum);
	}

    do 
    {
        EvtGetEvent(&event, evtWaitForever);
        
        if (event.eType==penDownEvent)
        {
        	PenGetPoint(&penX, &penY, &penDown);
			objectCode = TranslatePenToObjectButton(penX,penY);
			if (objectCode != noObject)
			{
				DepressObjectLSButton(objectCode);
				objectReleasePending = objectCode;
				continue;
			}
			else
			{
				objectReleasePending = noObject;
			}
		}
        if (event.eType==penUpEvent)
		{
			if (objectReleasePending != noObject)
			{
				PenGetPoint(&penX, &penY, &penDown);
				objectCode = TranslatePenToObjectButton(penX,penY);
				ReleaseObjectLSButton(objectReleasePending);
					
				if (objectReleasePending == objectCode)
				{
					switch(objectReleasePending)
					{
						case objectLoad:
							statusObjectLSActive=1;
							UpdateObjectLoadOrSaveTab(saveFieldIndex,false);
							break;
						case objectSave:
							statusObjectLSActive=2;
							UpdateObjectLoadOrSaveTab(saveFieldIndex,false);
							break;
						case objectCancel:
							applyChangesOrCancel=1;
							break;
						case objectOK:
							applyChangesOrCancel=2;
							break;
						default: break;
					}
					continue;
				}
			}
		}
		
		if (event.eType==appStopEvent)  // respond to app-switching events.
		{
			DepressObjectLSButton(objectCancel);
			SysTaskDelay(10);
			ReleaseObjectLSButton(objectCancel);
			statusQuit=true;
			applyChangesOrCancel=1;
			EvtAddEventToQueue(&event);  // put app stop back on queue
			continue;
		}

		if (event.eType==keyDownEvent)
		{
			if (event.data.keyDown.chr == chrLineFeed)
			{
				DepressObjectLSButton(objectOK);
				SysTaskDelay(10);
				ReleaseObjectLSButton(objectOK);
				applyChangesOrCancel=2;
				continue;
			}
		}
		
	 	if (event.eType==volumeUnmountedEvent)
		{
			if (event.data.generic.datum[0]==msVolRefNum)
			{
				statusInUnmountDialog = true;
				if (OpenAlertConfirmDialog(alert,pleaseReinsertCardString,true,false))
					statusVolumeMounted=false;
				else
					statusVolumeMounted=true;
				statusInUnmountDialog = false;
			}
			continue;
		}
		
		if (event.eType==volumeMountedEvent)
		{
			CheckForRemount();
			continue;
		}
		
		if (event.eType==displayChangeEvent)
		{
			continue;
		}

		if (! SysHandleEvent(&event))
        {
			FrmDispatchEvent(&event);
        }
    } 
    while (!applyChangesOrCancel);

	FldReleaseFocus(saveFileField);     // use FldSetUsable & FldReleaseFocus instead of FrmHideObject
	FldSetUsable(saveFileField,false);  // I don't like how FrmHideObject erases...
	
exitObjectLS:
	
	EvtFlushPenQueue();
	
	if (applyChangesOrCancel==2)
	{
		if (statusObjectLSActive==1)
		{
			LoadObject(fileList[selectedFile]);
			simulate_keypress(keyON);
		}
		else
		{
   			saveFilePtr = MemHandleLock(saveFileHandle);
   			for (j=0;j<numFiles;j++)
			{
				if (StrCompare(saveFilePtr,dirFileNames[j])==0) break;
			}
			if (j!=numFiles)
			{
				if (OpenAlertConfirmDialog(confirm,fileExistsString,true,false))
				{
					SaveObject(saveFilePtr);
				}
				else
				{
					applyChangesOrCancel=0;
					goto UserInputLoop;
				}
			}
    		SaveObject(saveFilePtr);
   			MemHandleUnlock(saveFileHandle);
		}
	}
	
	for (fileIndex=numFiles-1;fileIndex>=0;fileIndex--)
		MemPtrFree(dirFileNames[fileIndex]);
	FldSetTextHandle( saveFileField, NULL );
	MemHandleFree(saveFileHandle);
	
	if (!option320x320)
	{
		SilkLibEnableResize(silkLibRefNum);
		SilkLibResizeDispWin(silkLibRefNum, silkResizeMax);  // bring up graffiti area
	}
	
	PixelDissolve(backStoreWindow,mainDisplayWindow,320L*objectLSDialogExtentY,objectLSDialogOriginY,objectLSDialogOriginY,false);
}
